/*
 * Copyright (c) 1997, Purdue University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that:
 *
 * (1) source distributions retain this entire copyright notice and comment,
 *     and
 * (2) distributions including binaries display the following acknowledgement:
 *
 *        "This product includes software developed by Purdue University."
 *
 *     in the documentation or other materials provided with the distribution
 *     and in all advertising materials mentioning features or use of this
 *     software.
 *
 * The name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * TCLLOG was developed by:
 *
 *	Todd M. Helfter
 *
 *	Purdue University
 *	Computing Center
 *	1408 Math Sciences Building
 *	West Lafayette, Indiana 47907-1408 U.S.A.
 *
 * For more information on TCLLOG, to report problems or to make comments,
 * send electronic mail to:
 *
 *	tmh@purdue.edu
 */

/*
 * tcllog.c
 */

#include "tcllog.h"

/*
 * Tcl_Invoke --
 * Call this somewhat like Tcl_VarEval:
 * Tcl_Invoke(interp, cmdName, arg1, arg2, ..., NULL);
 * Each arg becomes one argument to the command,
 * with no substitutions or parsing.
 *
 * Courtesy of Brent Welch's Book "Practical Programming in Tcl and Tk"
 *
 * This function has been modified from it's original.
 */
int Tcl_Invoke(va_alist)
	va_dcl				/* Variable number of arguments */
{
	Tcl_Interp *interp;
	char *cmd;			/* Command name */
	char *arg;			/* Command argument */
	char **argv;			/* String vector for arguments */
	int argc, i, max;		/* Count of arguments */
	Tcl_CmdInfo info;		/* Info about command procedures */
	va_list pvar;			/* varargs stuff */
	int result;			/* TCL_OK or TCL_ERROR */

	va_start(pvar);
	interp = va_arg(pvar, Tcl_Interp *);
	cmd = va_arg(pvar, char *);
	/*
	 * Map from the command name to a C procedure.
	 */
	if (! Tcl_GetCommandInfo(interp, cmd, &info)) {
		Tcl_AppendResult(interp, "Unknown command \"",
			cmd, "\"", NULL);
		va_end(pvar);
		return TCL_ERROR;
	}
	max = 10;
	argc = 1;
	argv = (char **)Tcl_Alloc(max * sizeof(char *));
	argv[0] = cmd;
	Tcl_ResetResult(interp);

	/*
	 * Build a vector out of the rest of the arguments.
	 */
	while (1) {
		arg = va_arg(pvar, char *);
		argv[argc] = arg;
		if (arg == (char *)NULL) {
			break;
		}
		argc++;
		if (argc >= max) {
			/*
			 * Allocate a bigger vector and copy old values in.
			 */
			char **old = argv;
			argv = (char **)Tcl_Alloc(2*max * sizeof(char *));
			for (i=0 ; i<max ; i++) {
				argv[i] = old[i];
			}
			free((char *)old);
			max = 2*max;
		}
	}
	va_end(pvar);
	/*
	 * Invoke the C procedure.
	 */
	result = (*info.proc)(info.clientData, interp, argc, argv);
	/*
	 * Release our references to the arguments. 
	 */
	Tcl_Free((char *)argv);
	return result;
}


/*
 * Tcllog_Init 
 */
int Tcllog_Init (interp)
	Tcl_Interp *interp;
{
	char buf[100];
	static char *LogInfoArray = "log_info";
	static char *LogUpToArray = "log_upto";
	static char *LogMaskArray = "log_mask";

	/*	
	 * set up the syslog facility codes
	 */
	sprintf(buf, "%d", LOG_KERN);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_KERN", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_USER);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_USER", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_MAIL);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_MAIL", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_DAEMON);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_DAEMON", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_AUTH);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_AUTH", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_SYSLOG);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_SYSLOG", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LPR);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LPR", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_NEWS);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_NEWS", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_UUCP);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_UUCP", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_CRON);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_CRON", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL0);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL0", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL1);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL1", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL2);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL2", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL3);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL3", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL4);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL4", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL5);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL5", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL6);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL6", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_LOCAL7);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_LOCAL7", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(facList)", buf, NULL);

	sprintf(buf,"%d",LOG_NFACILITIES);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_NFACILITIES", buf, TCL_GLOBAL_ONLY);
	sprintf(buf,"%d",LOG_FACMASK);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_FACMASK", buf, TCL_GLOBAL_ONLY);

	/*	
	 * set up the syslog priority codes
	 */
	sprintf(buf,"%d",LOG_EMERG);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_EMERG", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_ALERT);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_ALERT", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_CRIT);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_CRIT", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_ERR);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_ERR", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_WARNING);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_WARNING", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_NOTICE);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_NOTICE", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_INFO);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_INFO", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_DEBUG);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_DEBUG", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(ptyList)", buf, NULL);

	sprintf(buf,"%d",LOG_PRIMASK);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_PRIMASK", buf, TCL_GLOBAL_ONLY);

	/*	
	 * set up the syslog option flags for openlog
	 */
	sprintf(buf,"%d",LOG_PID);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_PID", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(optList)", buf, NULL);

	sprintf(buf,"%d",LOG_CONS);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_CONS", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(optList)", buf, NULL);

	sprintf(buf,"%d",LOG_ODELAY);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_ODELAY", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(optList)", buf, NULL);

	sprintf(buf,"%d",LOG_NDELAY);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_NDELAY", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(optList)", buf, NULL);

	sprintf(buf,"%d",LOG_NOWAIT);
	Tcl_SetVar2(interp, LogInfoArray, "LOG_NOWAIT", buf, TCL_GLOBAL_ONLY);
	Tcl_Invoke(interp, "lappend", "syslog(optList)", buf, NULL);


	/*	
	 * set up the syslog (LOG_UPTO) priorities for setlogmask
	 */
	sprintf(buf,"%d",LOG_UPTO(LOG_EMERG));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_EMERG", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_ALERT));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_ALERT", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_CRIT));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_CRIT", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_ERR));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_ERR", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_WARNING));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_WARNING", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_NOTICE));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_NOTICE", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_INFO));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_INFO", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_UPTO(LOG_DEBUG));
	Tcl_SetVar2(interp, LogUpToArray, "LOG_DEBUG", buf, TCL_GLOBAL_ONLY);



	/*	
	 * set up the syslog (LOG_MASK) priorities for setlogmask
	 */
	sprintf(buf,"%d",LOG_MASK(LOG_EMERG));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_EMERG", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_ALERT));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_ALERT", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_CRIT));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_CRIT", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_ERR));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_ERR", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_WARNING));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_WARNING", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_NOTICE));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_NOTICE", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_INFO));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_INFO", buf, TCL_GLOBAL_ONLY);

	sprintf(buf,"%d",LOG_MASK(LOG_DEBUG));
	Tcl_SetVar2(interp, LogMaskArray, "LOG_DEBUG", buf, TCL_GLOBAL_ONLY);


	/*	
	 * add the syslog commands to the interpreter
	 */
        Tcl_CreateCommand(interp, "openlog", open_tcllog,
                (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
        Tcl_CreateCommand(interp, "syslog", write_tcllog,
                (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
        Tcl_CreateCommand(interp, "closelog", close_tcllog,
                (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
        Tcl_CreateCommand(interp, "setlogmask", mask_tcllog,
                (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	/*	
	 * provide all of this as a package.
	 */
	if (Tcl_PkgProvide(interp, "Tcllog", TCLLOG_VERSION) != TCL_OK) {
		return TCL_ERROR;
	}
	return TCL_OK;
}


/*
 * open_tcllog() -- 
 */
int open_tcllog(clientData, interp, argc, argv)
	ClientData      clientData;
	Tcl_Interp      *interp;
	int             argc;
	char		**argv;
{
	char buf[100];
	int err, opt, fac;

	Tcl_ResetResult(interp);
	if (argc < 4) {
		Tcl_AppendResult(interp, argv[0], ": wrong # args.", (char*) NULL);
		return TCL_ERROR;
	}


	if ((err = Tcl_GetInt(interp, argv[2], &opt)) != TCL_OK) {
		return TCL_ERROR;
	}
	sprintf(buf, "lsearch -exact $syslog(optList) %d", opt);
	Tcl_GlobalEval(interp, buf);
	if (atoi(interp->result) < 0) {
		Tcl_AppendResult(interp, "openlog: invalid option flag", (char*) NULL);
		return TCL_ERROR;
	}

	if ((err = Tcl_GetInt(interp, argv[3], &fac)) != TCL_OK) {
		return TCL_ERROR;
	}
	sprintf(buf, "lsearch -exact $syslog(facList) %d", fac);
	Tcl_GlobalEval(interp, buf);
	if (atoi(interp->result) < 0) {
		Tcl_AppendResult(interp, "openlog: invalid facility", (char*) NULL);
		return TCL_ERROR;
	}

	(void) openlog(argv[1], opt, fac);
	return TCL_OK;
}


/*
 * write_tcllog() -- Function which enables Tcl code to syslog at the
 *                   LOG_DEBUG level when needed.
 */
int write_tcllog(clientData, interp, argc, argv)
	ClientData      clientData;
	Tcl_Interp      *interp;
	int             argc;
	char		**argv;
{
	char buf[100];
	int pty;

	pty = atoi(argv[1]);
	sprintf(buf, "lsearch -exact $syslog(ptyList) %d", pty);
	Tcl_GlobalEval(interp, buf);
	if (atoi(interp->result) < 0) {
		Tcl_SetResult(interp, "openlog: invalid priority", TCL_STATIC);
		return TCL_ERROR;
	}

        (void) syslog(pty, argv[2]);
        return TCL_OK;
}


/*
 * close_tcllog() -- 
 */
int close_tcllog(clientData, interp, argc, argv)
	ClientData      clientData;
	Tcl_Interp      *interp;
	int             argc;
	char		**argv;
{
        (void) closelog();
        return TCL_OK;
}


/*
 * mask_tcllog() -- 
 */
int mask_tcllog(clientData, interp, argc, argv)
	ClientData      clientData;
	Tcl_Interp      *interp;
	int             argc;
	char		**argv;
{
	int err, msk;

	Tcl_ResetResult(interp);
	if (argc != 2) {
		Tcl_AppendResult(interp, argv[0], ": wrong # args.", (char*) NULL);
		return TCL_ERROR;
	}

	if ((err = Tcl_GetInt(interp, argv[1], &msk)) != TCL_OK) {
		return TCL_ERROR;
	}
	err = setlogmask(msk);
	return TCL_OK;
}
